/*
 * AbstractTextWrap.js
 *
 * Abstract text wrapping
 *
 * Copyright - Lombardi Software 2005
 * 
 * @version 0.1 - 03.09.2005 
 * @author Hugh Leather (hughleat@hotmail.com) 
 */

 /**
 * This class is used to do text wrapping.  As an abstract class, derivees 
 * should ensure to provide suitable implementations of the abstract methods.
 * The functions are 'getTextWidth' and 'outputLine'. 
 * <br/>
 * The wrapping happens inside an abstract rectangle whose dimensions are 
 * supplied.  Many different options are available for controlling both the ways
 * that text is split at line boundaries and the way layout is performed.
 * <br/>
 * Once an object is properly instantiated, call the 'wrap()' method to do the
 * actual wrapping.  Alternatively, the text can be only split into lines with
 * the 'split()' method and lines can be laid out with the 'layout()' method.
 * 
 * @param string the text to wrap - "" if not specified (from prototype)
 * @param width the width of the box to wrap in; this is not in characters but 
 * 		'user units', the meaning of which is supplied by overriding definitions 
 * 		abstract methods - 100 if not specified (from prototype)
 * @param height the height of the box to wrap in; this is not in characters but 
 * 		'user units', the meaning of which is supplied by overriding definitions 
 * 		abstract methods - 100 if not specified (from prototype)
 * @param fontSize the height of the the font at scale of 1.0 - 1 if not 
 *      specified (from prototype)
 * @param lineInterval the distance between two lines (top of line to top of 
 *      line) at scale of 1.0 - 1 if not specified (from prototype)
 * @param hFill horizontal fill policy; this should be one of the constants in
 *		AbstractTextWrap.hFill; see comments there for an explanation values 
 *		meanings - SHRINK if not specified (from prototype)
 * @param vFill vertical fill policy; this should be one of the constants in
 *		AbstractTextWrap.vFill; see comments there for an explanation values 
 *		meanings - SHRINK if not specified (from prototype)
 * @param hAlign horizontal alignment policy; this should be one of the 
 *		constants in AbstractTextWrap.hAlign; see comments there for an 
 *		explanation values meanings - CENTER if not specified (from prototype)
 * @param vAlign vertical alignment policy; this should be one of the 
 *		constants in AbstractTextWrap.vAlign; see comments there for an 
 *		explanation values meanings - CENTER if not specified (from prototype)
 * @param wordSep a regular expression used to split the string into individual
 *		words - /\s+\ (i.e. words are separated by at least one	whitespace 
 *		character) if not specified (from prototype); must be a regular 
 *      expression
 * @param hyphen if the hFill mode is HYPHENATE and a word is split at the end
 *		of a line, then this is the hyphen string to use - "-" if not specified
 *		(from prototype)
 * @param continuation if the cFill mode is CLIP and there are too many lines to
 *		fit in the text box, then this continuation string will be placed at the
 *		end of the fitting text to indicate text has been truncated - "..."
 *		if not specified (from prototype)
 */ 
function AbstractTextWrap( 
	string, 
	width, height, fontSize, lineInterval,
	hFill, vFill,
	hAlign, vAlign,
	wordSep, hyphen, continuation
) {
	if( string != null ) this._string = string;
	if( width != null ) this._width = width;
	if( height != null ) this._height = height;
	if( fontSize != null ) this._fontSize = fontSize;
	if( lineInterval != null ) this._lineInterval = lineInterval;

	if( hFill != null ) this._hFill = hFill;
	if( vFill != null ) this._vFill = vFill;
	if( hAlign != null ) this._hAlign = hAlign;
	if( vAlign != null ) this._vAlign = vAlign;

	if( wordSep != null ) this._wordSep = wordSep
	if( hyphen != null ) this._hyphen = hyphen;
	if( continuation != null ) this._continuation = continuation;
}
/**
 * Constants for horizontal fill policy.  This determines how the algorithm 
 * responds to lines not fitting horizontally.  For a line not to fit, the 
 * smallest unit (a word as separated by wordSep) must be too big for a line by
 * itself, otherwise the wrapping would have simply wrapped the word over to the
 * next line.
 * <br/>
 * So, when the algorithm cannot fit a single word on a line by itself, it takes
 * action based on the value of its hFill paramter.  This is one of the 
 * constants below.  
 * <br/>
 *   SHRINK : the scale of the text is reduced so that the line will fit and
 *   wrapping begins again.
 *   EXPAND : the word simply extends outside of the text box.
 *   HYPHENATE : the word is broken in two by the getHyphenPosition() method
 *   and the first part (with the hyphen string appended) is attempted to be 
 *   wrapped again (the remainder is pushed over to the next line).  If not even
 *   one character and the hyphen are possible on the line, then no text is 
 *   output at all, the rational being that the text box is too small to fit 
 *   anything useful (I tell a lie, actually, only the continuation is printed).
 */
AbstractTextWrap.hFill = { 
	SHRINK : "shrink", 
	EXPAND : "expand", 
	HYPHENATE : "hyphenate" 
};
/**
 * Constants for vertical fill policy.  This determines how the algorithm 
 * responds to lines not fitting vertically.  Once horizontal splitting has been
 * done, there may be too many lines to fit in the height of the text box.
 * <br/>
 * So, when the algorithm cannot fit a all the lines in the text box, it takes
 * action based on the value of its vFill paramter.  This is one of the 
 * constants below.  
 * <br/>
 *   SHRINK : the scale of the text is reduced until the text fits.  This is 
 *   done by contiually retrying at different scales until a good fit is found.
 *   EXPAND : the lines simply extend outside of the text box.
 *   CLIP : the lines are truncated.  The algorithm will replace the last few
 *   letters of the last line with the continuation string.  If that string is
 *   alone too wide to place on the line according to the horizontal fill policy
 *   then it is placed there regardless.
 */
AbstractTextWrap.vFill = { 
	SHRINK : "shrink", 
	EXPAND : "expand", 
	CLIP : "clip" 
};
/**
 * Constants for the horizontal alignment policy.  This determines how the lines
 * will be laid out in the text box.  The meanings of these constants should be
 * immediately apparent to any who have used a word processor before.
 * <br/>
 *   LEFT : text is aligned to the left side of the box.
 *   CENTER : text is aligned to the middle of the box.
 *   RIGHT : text is aligned to the right side of the box.
 *   JUSTIFY : interword spacing is adjusted so that multi-word lines fit from
 *   one side of the box to the other.  Note that since there is no concept of a
 *   paragraph, this is applied even to the last line of the text.
 */
AbstractTextWrap.hAlign = { 
	LEFT : "left", 
	CENTER : "center", 
	RIGHT : "right",
	JUSTIFY : "justify"
};
/**
 * Constants for the vertical alignment policy.  This determines how the lines
 * will be laid out in the text box.  The meanings of these constants should be
 * immediately apparent to any who have used a word processor before.
 * <br/>
 *   TOP : text is aligned to the top side of the box.
 *   CENTER : text is aligned to the middle of the box.
 *   BOTTOM : text is aligned to the bottom side of the box.
 */
AbstractTextWrap.vAlign = { 
	TOP : "top", 
	CENTER : "center", 
	BOTTOM : "bottom"
};

/**
 * This method computes the length of a piece of text.  It is abstract and 
 * should be overriden by derived classes.  The meaning of the units for the
 * returned value should be determined by derived classes.
 * 
 * @param string the text to get the length of
 * @param fontSize the size of the font; if either of the fill policies selected
 * was SHRINK, then it the text may have been scaled to fit.  Thus, this value 
 * is no more than the fontSize passed into the constructor, but may be less.
 * @return a number indicating the width of the text; must always return a 
 * positive, non-zero number if the string has non-zero length and the scale is
 * greater than zero and the width of the text box is non zero.
 */
AbstractTextWrap.prototype.getTextWidth = function( string, fontSize ) {
	throw new Error( "Call to abstract method, AbstractTextWrap.prototype.getTextWidth()" );
}

/**
 * This method is called during layout to output a line.
 * 
 * @param line the text to output next
 * @param x the x coordinate to draw text at
 * @param y the y coordinate to draw text at
 * @param fontSize the size of the font; if either of the fill policies selected
 * was SHRINK, then it the text may have been scaled to fit.  Thus, this value 
 * is no more than the fontSize passed into the constructor, but may be less.
 * @param wordSpacing the amount of extra space to add between each word; used
 * for horizontal justification; will be undefined or null if not justifying.
 */
AbstractTextWrap.prototype.outputLine = function( line, x, y, fontSize, wordSpacing ) {
	throw new Error( "Call to abstract method, AbstractTextWrap.prototype.appendLine()" );
}
/**
 * Get the best hyphen position for a word.
 * <br/>
 * The word passed in may have been created from a larger word which had to be
 * hyphenated.  For example, if 'hyphenated' had previously been hyphenated at
 * position 2, giving 'hy-phenated' then this function would likely be called
 * with 'phenated' at a later time.
 * <br/>
 * The method should return a number between 0 (inclusive) and limit (exclusive)
 * <br/>
 * The meaning of the limit is that a hyphen should not be placed on or after 
 * that point.  This is typical when the text wrapper has attempted to hyphenate
 * at a first choice and found that string too long.  For example, suppose the 
 * word, 'hyphenated', could not fit on the line and was to be split, and this 
 * method prefered to split it at 'hyphenat-ed'.  Now, initially this method 
 * would be called with limit = 'hyphenated'.length and to split at the point
 * noted, would return 'hyphenat'.length.  If the text wrapper cannot fit the
 * new, shorter word ('hyphenat-'), then it will ask for a shorter hyphen 
 * position.  To do this, it will call with limit = 'hyphenat'.length.  In this
 * way the preferred position may be tried and then successivly shorter ones if
 * the preference fails.
 * <br/>
 * Returning zero, then means that the word cannot be split.
 * <br/>
 * By default, this method simply returns the floor of limit/2.
 *
 * @param word the word to be split
 * @param limit one more than the last possible hyphen position; will never be
 * more than word.length and never less than one.
 * @return the position to hyphenate at; must be in the range [0, limit).
 */
AbstractTextWrap.prototype.getHyphenPosition = function( word, limit ) {
	return Math.floor( limit / 2 );
}

/**
 * Split the text according to the parameters laid out in the constructor.
 *
 * @return an object with two fields, 'lines' and 'scale'; the former is the
 * splitting up of the lines, the latter is the scale required 
 */
AbstractTextWrap.prototype.split = function() {
	// Do some simple checks for the text box having zero area
	if( this._width == 0 && this._hFill == AbstractTextWrap.hFill.SHRINK ) {
		return { lines : [], scale : 0 };
	}		
	if( this._height == 0 && this._vFill == AbstractTextWrap.vFill.SHRINK ) {
		return { lines : [], scale : 0 };
	}		

	/**
	 * Binary search.
	 *
	 * Finds the last index, i, in [low,high) such that pred(i,data) = true
	 * The predicate results must be sorted:
	 *  	pred(i,data)=false => pred(i+1,data)=false for all i in N and
	 *  	pred(i,data)=true  => pred(i-1,data)=true  for all i in N and
	 * (i.e. all typical binary search type stuff.)
	 * If no such index is found in the range, then low-1 is returned.
	 *
	 * @param low the low index
	 * @param high the high index must be > low
	 * @param pred a function taking an integer and the data
	 * @param data data
	 * @return the index found
	 */
	function binarySearch( low, high, pred, data ) {
		function bsearch( low, high ) {
			var index;
			if( high - low <= 1 ) return low;
			else {
				var mid = (( high - low ) >> 1 ) + low;
				if( pred( mid, data )) {
					index = bsearch( mid, high );		
				} else {
					index = bsearch( low, mid );
				}
			}
			return index;
		}
		return bsearch( low - 1, high );
	}
	// Split the words up according to the word separator regexp
	var words = this._string.split( this._wordSep );
	
	// Start with full height text
	var scale = 1.0;
	// If the hFill policy is SHRINK, then we should find a scale that fits the
	// longest word.  This means that the hSplit function below does not need to
	// be recursive.
	if( this._hFill == AbstractTextWrap.hFill.SHRINK ) {
		for( var i = 0; i < words.length; ++i ) {
			if( words[ i ].length == 0 ) continue;
			var s = this._width / this.getTextWidth( words[ i ], this._fontSize );
			if( s < scale ) scale = s;
		}
	}
	
	/**
	 * A function to do the horizonal splitting based a given scale.
	 *
	 * @param the words to split across lines
	 * @param the scale to split at
	 * @param maxLines if defined, then attempt to layout at most this number of
	 * lines
	 * @return an array of lines.
	 * @throws the list of lines currently fitted if maxLines is defined, it 
	 * needs to add a new line and that would make more lines than maxLines.
	 */
	function hSplit( words, scale, maxLines ) {
		// An array to hold our lines as we construct them
		var lines = new Array();
		// The current line
		var line = null;
		
		/**
		 * A tiny helper function to ensure that whenever a line is added, the
		 * current line is reset.
		 * Additionally, it checks that we aren't adding too many lines and 
		 * throws the current line list if we are
		 *
		 * @param lineToAdd the line to add
		 * @throws the line list if too many lines
		 */
		function addLine( lineToAdd ) {
			if( maxLines != null && lines.length == maxLines ) {
				throw lines;
			}
				
			lines.push( lineToAdd );
			line = null;
		}
		
		// Keep picking off words trying to put them on the line
		while( words.length != 0 ) {
			// If this the first word of the line, this variable will be true
			var isFirstWord = line == null;
			// Take the next word
		    var word = words.shift();
		    // This will be the line we will try to fit
		    var candidate = isFirstWord ? word : line + " " + word;
		    // this is the width of the candidate line
		    var w = this.getTextWidth( candidate, scale * this._fontSize );
		
			if( w > this._width ) {
				// This word could not be fitted
				if( isFirstWord ) {
					// If we couldn't fit the first word, then we either have to
					// hyphenate it or just add it any way.  Note, that if we
					// are in SHRINK mode, then we must always be able to fit 
					// the first word because we've already chosen the scale to
					// do exactly that.
					if( this._hFill == AbstractTextWrap.hFill.HYPHENATE ) {
						// Find a hyphen position that fits
						for( var p = this.getHyphenPosition( word, word.length );
							 p > 0; p = this.getHyphenPosition( word, p )) {

							// This is what is on the left of the split
		                    var left = word.substring( 0, p ) + this._hyphen;
		                    if( this.getTextWidth( left, scale * this._fontSize ) <= this._width ) {
		                    	// This can be fitted.
		                        addLine( left );
		                        // Put back the remainder of the word to the 
		                        // next line
		                        words.unshift( word.substring( p ));
		                        break;
		                    }
						}
	                    if( p == 0 ) {
	                        // Can't even manage one lousy letter.
	                        // Fail miserably and just put out the 
	                        // continuation
	                        lines.length = 0;
	                        addLine( this._continuation );
	                        return lines;
	                    }
					} else {
						// Must be in EXPAND mode, just add the word anyway
						addLine( candidate );
					}
				} else {
					// Push it back to the next line and commit this line
					words.unshift( word );
					addLine( line );
				}
			} else { 
				// The line could be fitted
				line = candidate;
			}
		}
		// Make sure we add the last line
		if( line != null ) addLine( line );
		return lines;
	}
	
	/**
	 * Gte the scale needed to fit a number of lines into the height
	 */
	function getScaleToFitNumLines( numLines ) {
		return this._height / ( this._fontSize + ( numLines - 1 ) * this._lineInterval );
	}
	/**
	 * Get the maximum number of lines 
	 */
	function getMaxLinesForScale( scale ) {
		var h = this._height - this._fontSize * scale;
		return ( h < 0 ) ? 0 : Math.floor( h / ( this._lineInterval * scale )) + 1;
	}
	
	// Do the vertical split (which calls hSplit)
	var lines;
	switch( this._vFill ) {
	default :
	case AbstractTextWrap.vFill.EXPAND :
		// If we're just expanding, then we don't care how many lines we get
		lines = hSplit.call( this, words, scale );
		break;
	case AbstractTextWrap.vFill.CLIP :
		// In clip mode, we only need so many lines
		var maxLines = getMaxLinesForScale.call( this, scale, this._height );
		if( maxLines == 0 ) {
			// If we can't do any lines, then always return the continuation
		    lines = [ this._continuation ];
		} else {
			var thrownLines;
			try {
				// If this works, then we could fit all the lines
				lines = hSplit.call( this, words, scale, maxLines );
			} catch( thrownLines ) {
				// Otherwise we have to clip
				lines = thrownLines;
		        var last = lines[ lines.length - 1 ];
		        function pred( i, textWrap ) {
		        	var cont = last.substring( 0, i ) + textWrap._continuation;
		        	return textWrap.getTextWidth( cont, scale * this._fontSize ) <= textWrap._width;
		        }
		        last = last.substring( 0, binarySearch( 0, last.length, pred, this )) + this._continuation;
		        lines[ lines.length - 1 ] = last;
			}
		}
		break;
	case AbstractTextWrap.vFill.SHRINK :
		// If we are shrinking vertically, first try to see if it can be done in
		// the height allowed.
		lines = hSplit.call( this, [].concat( words ), scale );
		var low = getMaxLinesForScale.call( this, scale, this._height );
		if( lines.length > low ) {
			// If not then do a binary search to find the scale required.
			function pred( i, textWrap ) {
				var s = getScaleToFitNumLines.call( textWrap, i, textWrap._height );
				try {
					lines = hSplit.call( textWrap, [].concat( words ), s, i );
					return false;
				} catch( ignore ) {
					return true;
				}
			}
			var high = lines.length;			
			var numLines = binarySearch( low, high, pred, this ) + 1;
			scale = getScaleToFitNumLines.call( this, numLines, this._height );		
			lines = hSplit.call( this, words, scale );
		}
        break;
	}
	
	// Return the results.
	return { lines : lines, scale : scale };
}

/**
 * Layout the lines as were split from the split method.  Where the split method
 * uses the horizontal and vertical fill fields, this uses the horizontal and
 * vertical aligments. 
 *
 * @param lines an array of strings to be laid out
 * @param scale amount the text had to be scaled to fit into the textbox 
 */
AbstractTextWrap.prototype.layout = function( lines, scale ) {
	// First compute the widths for each line
	var widths = new Array();
	for( var i = 0; i < lines.length; ++i ) {
		widths[ i ] = this.getTextWidth( lines[ i ], scale * this._fontSize );
	}

	/**
	 * A quick function to get the height of a number of lines of text
	 * 
	 * @param numLines the number of lines
	 * @return the height
	 */
	function getTextHeight( numLines, scale ) {
		if( numLines == 0 ) return 0;
		return ( this._fontSize + this._lineInterval * ( numLines - 1 )) * scale;
	}

	// Determine where the top of the text should start relative to the top-
	// left corner of the text box (positive y grows down, x grows right).
	var height = getTextHeight.call( this, lines.length, scale );
	var y = 0;
	switch( this._vAlign ) {
	case AbstractTextWrap.vAlign.TOP : break;
	case AbstractTextWrap.vAlign.BOTTOM : 
		y = this._height - height; 
		break;
	case AbstractTextWrap.vAlign.CENTER : 
		y = ( this._height - height ) / 2;
		break;
	}

	// Layout the lines
	for( var i = 0; i < lines.length; ++i ) {
		// Get the position of the left edge of the text, relative to the left
		// edge of the text box
		var x = 0;
		switch( this._hAlign ) {
		case AbstractTextWrap.hAlign.JUSTIFY : break;
		case AbstractTextWrap.hAlign.LEFT : break;
		case AbstractTextWrap.hAlign.CENTER : 
			x = ( this._width - widths[ i ] ) / 2;
			break;
		case AbstractTextWrap.hAlign.RIGHT :
			x = this._width - widths[ i ];
			break;
		}
		
		// If text is to be justified, calculate the word spacing
		var wordSpacing;
		if( this._hAlign == AbstractTextWrap.hAlign.JUSTIFY ) {
			var numWords = lines[ i ].split( this._wordSep ).length;			
			wordSpacing = ( this._width - widths[ i ] ) / ( numWords - 1 );
			// Leave the last line unjustified
			wordSpacing = ( i != lines.length - 1 ) ? wordSpacing : null;
		}

		this.outputLine( lines[ i ], x, y, this._fontSize * scale, wordSpacing );
		y += this._lineInterval * scale;
	}	
}

/**
 * Wrap the text.  First split the lines over the textbox, then lay them out.
 */
AbstractTextWrap.prototype.wrap = function() {
	var split = this.split();
	this.layout( split.lines, split.scale );
}

// Default values for the arguments to the constructor.
// If changes are made to these values, maintain the documentation in the 
// constructor.
AbstractTextWrap.prototype._string = "";
AbstractTextWrap.prototype._width = 100;
AbstractTextWrap.prototype._height = 100;
AbstractTextWrap.prototype._fontSize = 1;
AbstractTextWrap.prototype._lineInterval = 1;
AbstractTextWrap.prototype._hFill = AbstractTextWrap.hFill.SHRINK;
AbstractTextWrap.prototype._vFill = AbstractTextWrap.vFill.SHRINK;
AbstractTextWrap.prototype._hAlign = AbstractTextWrap.hAlign.CENTER;
AbstractTextWrap.prototype._vAlign = AbstractTextWrap.vAlign.CENTER;
AbstractTextWrap.prototype._wordSep = /\s+/;
AbstractTextWrap.prototype._hyphen = "-";
AbstractTextWrap.prototype._continuation = "...";

