/*
 * Flow.js
 *
 * Copyright - Lombardi Software 2005
 * 
 * @version 0.1 - 03.09.2005 
 * @author Hugh Leather (hughleat@hotmail.com) 
 */

/**
 * This class is used to transforms routes into paths.
 *
 * The static '_init' method is called to find all flows in the document and add
 * the appropriately routed paths.
 *
 * @param node the XML/SVG node describing the flow
 */ 
function Flow( node ) {
	this._node = node;
	this._parse();
	this._build();
	this._node.normalize();
}

/**
 * The namespace of the flow elements
 */
Flow.ns = "http://www.lombardi.com/bpd/svg/";
/**
 * The tag name
 */
Flow.tag = "flow";
/**
 * A list of instances that are found when '_init' is called
 */
Flow._instances = new Array();

/**
 * Find all the flow elements and route them.
 *
 * Since this moves text wraps around, that classes, init needs to be called
 * after this.
 */
Flow._init = function () {
	var elements = document.documentElement.getElementsByTagNameNS( this.ns, this.tag );
	for( var i = 0; i < elements.length; i++ ) {
		this._instances.push( new Flow( elements.item( i )));
	}
}

/**
 * Get the interesting data out of the node.  These parses the XML to get the
 * following fields (suitable for calling Vector.route):
 * _offset : number = offset from start or end to avoid flowObjects
 * _startP : Vector = first point in the flow
 * _startN : Vector = normal of first point
 * _ctrlPs : array of Vectors = intermediate points
 * _endP : Vector = last point in the flow
 * _endN : Vector = normal of the last point
 * _smooth : boolean = whether to have smooth flows or not
 * _textOnPath : boolean = whether to have text on paths or not
 */
Flow.prototype._parse = function() {
	this._offset = parseFloat( this._node.getAttribute( "offset" ));
	this._smooth = this._node.getAttribute( "smooth" ) == "true";
	this._textOnPath = this._node.getAttribute( "text-on-path" ) == "true";

	var from = this._node.getElementsByTagNameNS( Flow.ns, "from" ).item( 0 );
	this._startP = Vector.parseXMLElement( from.getElementsByTagNameNS( Flow.ns, "location" ).item( 0 ));
	this._startN = Vector.parseXMLElement( from.getElementsByTagNameNS( Flow.ns, "normal" ).item( 0 ));

	var controlPoints = this._node.getElementsByTagNameNS( Flow.ns, "controlPoint" );
	this._ctrlPs = new Array();
	for( var i = 0; i < controlPoints.length; ++i ) {
		this._ctrlPs.push( Vector.parseXMLElement( controlPoints.item( i )));
	}

	var to = this._node.getElementsByTagNameNS( Flow.ns, "to" ).item( 0 );
	this._endP = Vector.parseXMLElement( to.getElementsByTagNameNS( Flow.ns, "location" ).item( 0 ));
	this._endN = Vector.parseXMLElement( to.getElementsByTagNameNS( Flow.ns, "normal" ).item( 0 ));
	
	// Find out if there is a text node for this flow.
    // alert("this._node.getParentNode().getElementsByTagNameNS( TextWrap.ns, TextWrap.tag );	");
    // alert(this._node);	
    if (this._node.getParentNode) {
        var texts = this._node.getParentNode().getElementsByTagNameNS( TextWrap.ns, TextWrap.tag );	
        if( texts.length > 0 ) {
            this._text = texts.item( 0 );
        }
    }
}

/**
 * Insert the appropriately routed path.
 *
 * This will call Vector.route with the appropriate arguments and the points to
 * a path just after the flow node.  The path will be stored in 
 * _path
 */
Flow.prototype._build = function() {
	this._group = document.createElementNS( SVG.ns, "g" );

	var node = this._node;
	var nextElement = null;
	while( node.nextSibling ) {
		if( node.nextSibling.nodeType == 1 ) {
			nextElement = node.nextSibling;
			break;
		} else {
			node = node.nextSibling;
		}
	}
	
	if( nextElement ) {
		var test = this._node.parentNode.insertBefore( this._group, nextElement );
	} else {
		this._node.parentNode.appendChild( this._group );
	}
	
	var wayPoints = Vector.route( this._startP, this._startN, this._ctrlPs, this._endP, this._endN, this._offset );
	this._path = document.createElementNS( SVG.ns, "path" );
	this._group.appendChild( this._path );
	this._path.setAttribute( "style", this._node.getAttribute( "style" ));
	this._path.setAttribute( "id", "flow-id-" + Flow._instances.length );

	var path = "";
	function toPath( v ) { return v.x + "," + v.y + " "; }
	if( !this._smooth || wayPoints.length == 2 ) {
		path += "M" + toPath( wayPoints[ 0 ] );
		for( var i = 1; i < wayPoints.length; ++i ) {
			path += "L" + toPath( wayPoints[ i ] );
		}
	} else {
		path += "M" + toPath( wayPoints[ 0 ] );
		for( var i = 1; i < wayPoints.length - 2; ++i ) {
			path += "Q" + toPath( wayPoints[ i ] );
			with( Vector ) {
				var a = wayPoints[ i ], b = wayPoints[ i + 1 ];
				var mid = add( a, mul( sub( b, a ), 0.5 ));
				path += toPath( mid );
			}
		}
		path += "Q" + toPath( wayPoints[ wayPoints.length - 2 ] );
		path += toPath( wayPoints[ wayPoints.length - 1 ] );
	}	
	this._path.setAttribute( "d", path );	

	if( this._text ) {
		if( this._textOnPath ) {
			this._text.setAttribute( "on-path-url", "#" + this._path.getAttribute( "id" ));
		} else {
			this._positionTextAtMidPoint( wayPoints );
		}	
	}
}

/**
 * If there is text, determine the mid point of the flow and move the text to
 * be centered on that point.
 * 
 * @param wayPoints the list of point the flow will go through
 */
Flow.prototype._positionTextAtMidPoint = function( wayPoints ) {
	// Find the length of the flow
	var length = 0;
	for( var i = 0; i < wayPoints.length - 1; ++i ) {
		length += Vector.distance( wayPoints[ i ], wayPoints[ i + 1 ] );
	}
	// Find the segment that contains the mid point
	var midPoint;
	length /= 2;
	for( var i = 0; i < wayPoints.length - 1; ++i ) {
		var d = Vector.distance( wayPoints[ i ], wayPoints[ i + 1 ] );
		if( length <= d ) {
			// This segment contains the mid point
			midPoint = wayPoints[ i ];
			if( d != 0 ) {				
				midPoint = Vector.add( midPoint, Vector.mul( Vector.sub( wayPoints[ i + 1 ], wayPoints[ i ] ), length / d ));
			}
			break;
		}
		length -= d;
	}
	if( midPoint == null ) midPoint = wayPoints[ 0 ];
	
	// Move the text node to the right position
	var x = this._text.getAttribute( "x" );
	x = SVG.convertLengthToUserSpace( x, this._text.getParentNode( ));
	x += midPoint.x;
	this._text.setAttribute( "x", x );

	var y = this._text.getAttribute( "y" );
	y = SVG.convertLengthToUserSpace( y, this._text.getParentNode( ));
	y += midPoint.y;
	this._text.setAttribute( "y", y );
}
