/*
 * TextWrap.js
 *
 * Copyright - Lombardi Software 2005
 * 
 * @version 0.1 - 03.09.2005 
 * @author Hugh Leather (hughleat@hotmail.com) 
 */

/**
 * This class is used to do text wrapping
 *
 * The static '_init' method is called to find text wraps in the document and 
 * add the appropriately wrapped text.
 *
 * The class extracts information out of the SVG XML and calls the code in
 * AbtsractTextWrap.js to do the dirty work.
 *
 * @param node the XML/SVG node describing the text
 */ 
function TextWrap( node ) {
	this._node = node;
	this._node.normalize();

	// Get the parameters from the XML
	this._svg = SVG.addSVGSibling( node, document.createElementNS( SVG.ns, "text" ));
	this._svg.setAttribute( "class", "text" );
	this._svg.appendChild( document.createTextNode( "" ));
	this._svg.setAttribute( "style", this._node.getAttribute( "style" ));

	function onEmpty( x, def ) {
		return x == null || x == "" ? def : x;
	}

	this._x = onEmpty( node.getAttribute( "x" ), 0 );
	this._y = onEmpty( node.getAttribute( "y" ), 0 );
	var wordSep = onEmpty( getStyleProperty( this._svg,  "word-separator" ), null );
	if( wordSep != null ) wordSep = new RexExp( wordSep );
	AbstractTextWrap.call( this,
		this._node.firstChild != null ? this._node.firstChild.data : "",
		onEmpty( node.getAttribute( "width" ), null ),
		onEmpty( node.getAttribute( "height" ), null ),
		onEmpty( getStyleProperty( this._svg,  "font-size" ), null ),
		onEmpty( getStyleProperty( this._svg,  "line-interval" ), null ),
		onEmpty( getStyleProperty( this._svg,  "horizontal-fill" ), null ),
		onEmpty( getStyleProperty( this._svg,  "vertical-fill" ), null ),
		onEmpty( getStyleProperty( this._svg,  "horizontal-align" ), null ),
		onEmpty( getStyleProperty( this._svg,  "vertical-align" ), null ),
		wordSep,
		onEmpty( getStyleProperty( this._svg,  "hyphen" ), null ),
		onEmpty( getStyleProperty( this._svg,  "continuation" ), null )
	);
	
    if (node.getParentNode) {
        this._x = SVG.convertLengthToUserSpace( this._x, node.getParentNode( ));
        this._y = SVG.convertLengthToUserSpace( this._y, node.getParentNode( ));
        this._width = SVG.convertLengthToUserSpace( this._width, node.getParentNode( ));
        this._height = SVG.convertLengthToUserSpace( this._height, node.getParentNode( ));
        this._fontSize = SVG.convertLengthToUserSpace( this._fontSize, node.getParentNode( ));
    }
	this._svg.style.setProperty( "font-size", this._fontSize, "important" );
    if (this._svg.getParentNode) {
        this._lineInterval = SVG.convertLengthToUserSpace( this._lineInterval, this._svg.getParentNode() );
    }
	this._onPathUrl = node.getAttribute( "on-path-url" );
	if( this._onPathUrl == "" ) this._onPathUrl = null;
	
	var margin = node.getAttribute( "margin" );
	if( margin != null && margin != "" ) {
		margin = margin.split( /\s+|;|,/ );
		for( var i = 0; i < margin.length; ++i ) {
            if (this._svg.getParentNode) {
                margin[ i ] = SVG.convertLengthToUserSpace( margin[ i ], this._svg.getParentNode() );
            }
		}
		if( margin[ 0 ] != null ) {
			this._x += margin[ 0 ]; this._width -= margin[ 0 ];
		}
		if( margin[ 1 ] != null ) {
			this._y += margin[ 1 ]; this._height -= margin[ 1 ];
		}
		if( margin[ 2 ] != null ) {
			this._width -= margin[ 2 ];
		}
		if( margin[ 3 ] != null ) {
			this._height -= margin[ 3 ];
		}
	}

	this.wrap();
}

/**
 * Nasty hacky function because Batik doesn't let us see style properties it doesn't know about thru the style. object
 * We try and get the property from the style; if it is null we try parsing the "style" attribute directly.
 */
function getStyleProperty( node, propName)
{
	var style = node.style;
	var rc = style.getPropertyValue( propName);
	if (rc) return rc;
	var attr = node.getAttribute("style");
	var pos = attr.indexOf(propName);
	if (pos < 0) return null;
	pos += propName.length;
	var end = attr.length;
	var semi = attr.indexOf(';',pos);
	if (semi>0) end = semi;
	while (pos!=end && (attr.charAt(pos)==' ' || attr.charAt(pos)=='\t')) pos++;
	if (pos==end || attr.charAt(pos)!=':') return null;
	pos++;
	while (pos!=end && (attr.charAt(pos)==' ' || attr.charAt(pos)=='\t')) pos++;
	while (end>pos && (attr.charAt(end-1)==' ' || attr.charAt(end-1)=='\t')) end--;
	if (pos==end) return null;
	return attr.substring( pos, end);
}  

TextWrap.prototype = new AbstractTextWrap();
TextWrap.prototype._x = 0;
TextWrap.prototype._y = 0;
TextWrap.prototype._lineInterval = "1em";

/**
 * Override getTextWidth in AbstractTextWrap.
 * 
 * Delegates to the SVG function, getComputedTextLength.
 */
TextWrap.prototype.getTextWidth = function( string, fontSize ) {
	if( string.length == 0 ) return 0;
	this._svg.style.setProperty( "font-size", fontSize, "important" );
	this._svg.firstChild.data = string
	var width = this._svg.getComputedTextLength();	
	this._svg.firstChild.data = "";
	return width;
}

/**
 * Override outputLine in AbstractTextWrap.
 *
 * Creates a tspan to contain the line.  May also create an textPath if the
 * text should lie on some path.
 */
TextWrap.prototype.outputLine = function( line, x, y, fontSize, wordSpacing ) {
	this._svg.style.setProperty( "font-size", fontSize, "important" );
	var tspan = document.createElementNS( SVG.ns, "tspan" );
	tspan.appendChild( document.createTextNode( line )); 
	if( wordSpacing != null ) {
		tspan.style.setProperty( "word-spacing", wordSpacing, "important" );
	}
	tspan.setAttribute( "x", x + this._x );
	tspan.setAttribute( "y", y + this._y );
	tspan.setAttribute( "baseline-shift", "-100%" );
	
	if( this._onPathUrl != null ) {
		var textPath = document.createElementNS( SVG.ns, "textPath" );
		textPath.setAttributeNS( "http://www.w3.org/1999/xlink", "href", this._onPathUrl );
		textPath.setAttribute( "startOffset", "50%" );

		this._svg.appendChild( textPath );
		textPath.appendChild( tspan );		
	} else {
		this._svg.appendChild( tspan );
	}
}

/**
 * Namespace of text wrap elements
 */
TextWrap.ns = "http://www.lombardi.com/bpd/svg/";
/**
 * The tag name for this node
 */
TextWrap.tag = "text-wrap";

/**
 * A list of instances that are found when '_init' is called
 */
TextWrap._instances = new Array();

/**
 * Find all the text:wrap elements and wrap them.
 */
TextWrap._init = function () {
	var elements = document.documentElement.getElementsByTagNameNS( this.ns, this.tag );
	for( var i = 0; i < elements.length; i++ ) {
		this._instances.push( new TextWrap( elements.item( i )));
	}
}
