/*
 * Vector.js
 *
 * Copyright - Lombardi Software 2005
 * 
 * @version 0.1 - 03.09.2005 
 * @author Hugh Leather (hughleat@hotmail.com) 
 */

/**
 * This class represents simple two dimensional vectors.
 * <br> 
 * Object of this class have two public variables, x and y, with the obvious
 * meanings.
 *
 * @param x the x coordinate of the vector (must be a non-null number)
 * @param y the y coordinate of the vector (must be a non-null number)
 * @return the newly created vector
 */
function Vector( x, y ) {
	this.x = x;
	this.y = y;
}
/**
 * Create a vector from any xml node with two attributes, x and y.<b> 
 *
 * @param node the xml node, must not be null and must have attributes, x and y.
 * @return the newly created vector. 
 */
Vector.parseXMLElement = function( node ) {
	return new Vector( parseFloat( node.getAttribute( "x" )), parseFloat( node.getAttribute( "y" )));
}
/**
 * Compute a string representation of the vector.  Really just for debugging
 * purposes.
 * 
 * @return a string of the form '(x,y)' 
 */
Vector.prototype.toString = function() {
	return "(" + this.x + "," + this.y + ")";
}
/**
 * Add two vectors. 
 *
 * @param a first vector, may not be null
 * @param b second vector, may not be null
 * @return the vector, a + b
 */
Vector.add = function( a, b ) {
    return new Vector( a.x + b.x, a.y + b.y );
}
/**
 * Subtract two vectors.
 *
 * @param a first vector, may not be null
 * @param b second vector, may not be null
 * @return the vector, a - b
 */
Vector.sub = function( a, b ) {
	return new Vector( a.x - b.x, a.y - b.y );
}
/**
 * Multiply a vector by a scalar
 *
 * @param a first vector, may not be null
 * @param b scalar, may not be null
 * @return the vector, a * b
 */
Vector.mul = function( a, b ) {
	return new Vector( a.x * b, a.y * b );
}
/**
 * Get the scalar product of two vectors.
 *
 * @param a first vector, may not be null
 * @param b second vector, may not be null
 * @return the scalar, (a.b)
 */
Vector.dot = function( a, b ) {
	return a.x * b.x + a.y * b.y;
}
/**
 * Get the pseudo cross product of two vectors.
 *
 * @param a first vector, may not be null
 * @param b second vector, may not be null
 * @return the scalar, a.x*b.y - b.x*a.y
 */
Vector.cross = function( a, b ) {
	return a.x * b.y - b.x * a.y;
}
/**
 * Rotate a vector by 90 degrees, either clockwise or anticlockwise
 *
 * @param a vector, may not be null
 * @param clockwise true if clockwise, false if anticlockwise, may not be null
 * @return the vector, a, rotated by 90 degrees either clockwise or
 * anticlockwise.
 */
Vector.rotate90 = function( a, clockwise ) {
    return clockwise ? new Vector( a.y, -a.x ) : new Vector( -a.y, a.x );
}
/**
 * Get the magnitude of a vector
 *
 * @param a vector, may not be null
 * @return the scalar length of the vector
 */
Vector.magnitude = function( a ) {
	return Math.sqrt( Vector.dot( a, a ));
}
/**
 * Get the distance between two vectors
 *
 * @param a the first vector; may not be null
 * @param b the second vector; may not be null
 * @param the scalar distance between them
 */
Vector.distance = function( a, b ) {
	return Vector.magnitude( Vector.sub( b, a ));
} 	
/**
 * Get the distance squared between two vectors
 *
 * @param a the first vector; may not be null
 * @param b the second vector; may not be null
 * @param the scalar distance squared between them
 */
Vector.distance2 = function( a, b ) {
	var ab = Vector.sub( b, a );
	return ab.x * ab.x + ab.y * ab.y;
} 	
/**
 * Get the orthogonal distance squared between two vectors
 *
 * @param a the first vector; may not be null
 * @param b the second vector; may not be null
 * @param the scalar orthogonal distance between them
 */
Vector.distanceOrthogonal = function( a, b ) {
	var ab = Vector.sub( b, a );
	return Math.abs( ab.x ) + Math.abs( ab.y );
} 	
/**
 * Get the normalized version of a vector.<b>
 * <br>
 * If the vector is zero, then a new zero vector is returned 
 *
 * @param a vector, may not be null
 * @return the unit length normalized vector
 */
Vector.normalize = function( a ) {
	if( a.x == 0 && a.y == 0 ) return new Vector( 0, 0 );
	var len = Vector.magnitude( a );
	return new Vector( a.x / len, a.y / len );
}

/**
 * Find a route 
 *
 * @param startP first point; non-null vector
 * @param startN start normal; non-null vector
 * @param ctrlPs intermediate points; non-null array of non-null vectors
 * @param endP last point; non-null vector
 * @param endN last normal; non-null vector
 * @param offset the distance to avoid the start or end objects
 * @return the list of points in the route; an array of vectors
 */
Vector.route = function( startP, startN, ctrlPs, endP, endN, offset ) {
	// This will be the list of points in the route.
	var wayPoints = new Array();
	
	var Debug = {
		trace : function() {},
		displayTrace : function() {}
	};
	
	/**
	 * Route between two points.
	 *
	 * There may be many points in ctrlPs.  This function is called between
	 * every successive pair (including startP and endP).  
	 * 
	 * @param startP first point; non-null vector
	 * @param startN start normal; non-null vector
	 * @param isFirst true if this is the first point
	 * @param endP second point; non-null vector
	 * @param endN second normal; non-null vector
	 * @param isLast true if this is the last point
	 * @return the direction of the last segment
	 */
	function singleConnect( startP, startN, isFirst, endP, endN, isLast ) {
		Debug.trace( "singleConnect( " + startP + ", " + startN + ", " + isFirst + ", " + endP + ", " + endN + ", " + isLast + " )" );

		// A few oft used dot products
		var d = Vector.sub( endP, startP );
	    var ds = Vector.dot( startN, d );
	    var de = Vector.dot( endN, d );

		Debug.trace( "sC(1) d=" + d + " ds=" + ds + " de=" + de );		
	
		// If the start normal points away from where we going, then rotate by
		// 90 degrees toward the direction
		var firstN = ds > 0 ? startN : Vector.rotate90( startN, Vector.cross( startN, d ) <= 0 );
		// If this is the first point and the start normal points away from 
		// where we going, then 'push out' by offset and start again along the
		// rotated normal.  This drastically reduces the number of cases that
		// need to be considered.
		if( isFirst && ds <= 0 ) {
			var p = Vector.add( startP, Vector.mul( startN, offset ));
	        wayPoints.push( p );
		    var n = singleConnect( p, firstN, isFirst, endP, endN, isLast, offset );
	   		return n;
		}
	    
	    // Same deal for the end as for the start
	   	var lastN = de < 0 ? endN : Vector.rotate90( endN, Vector.cross( endN, d ) >= 0 );
		if( isLast && de >= 0 ) {
	        var p = Vector.add( endP, Vector.mul( endN, offset ));
	       	var n = singleConnect( startP, startN, isFirst, p, lastN, isLast, offset );
	        wayPoints.push( p );
	   	    return n;
		}

		Debug.trace( Vector.dot( startN, d ));
		Debug.trace( Vector.cross( startN, d ));
		if( Vector.cross( startN, d ) == 0 ) return startN;

		// If normals are orthogonal
	   	if( Vector.dot( firstN, lastN ) == 0 ) {
	   		if( Vector.dot( firstN, d ) >= 0 && Vector.dot( lastN, d ) <= 0 ) {
	   			var p = Vector.add( startP, Vector.mul( firstN, Vector.dot( firstN, d )));
	       		wayPoints.push( p );
	       		var n = Vector.normalize( Vector.sub( endP, p ));
	       		Debug.trace( "Orth " + n );
	            return n;
			} else {
				// Should not be possible to get here, cos we turned normals 
				// that pointed in the wrong direction.
	       		// Must return something.
	       		Debug.trace( "Vector.route.singleConnect - Unexpected state" );
			    return new Vector( 1, 0 );
			}
		} else {
			// Normals are parallel
	   		var p = Vector.add( startP, Vector.mul( firstN, 0.5 * Vector.dot( firstN, d )));
    	    var q = Vector.add( endP, Vector.mul( lastN, -0.5 * Vector.dot( lastN, d )));
   	    	wayPoints.push( p );
       		wayPoints.push( q );
       		var n = Vector.normalize( Vector.sub( endP, q ));
       		return n;
		}
	}

	// Do each pair in turn.
	wayPoints.push( startP );
	if( ctrlPs.length == 0 ) {
		singleConnect( startP, startN, true, endP, endN, true, offset );
	} else {
		var zero = new Vector( 0, 0 );
		Debug.trace( "Segment" + 0 );
		var n = singleConnect( startP, startN, true, ctrlPs[ 0 ], zero, false, offset );
		wayPoints.push( ctrlPs[ 0 ] );
		
		for( var i = 1; i < ctrlPs.length; ++i ) {
			Debug.trace( "Segment" + i );
			n = singleConnect( ctrlPs[ i - 1 ], n, false, ctrlPs[ i ], zero, false, offset );
			wayPoints.push( ctrlPs[ i ] );
		}
		// In order to match the existing Java code, the flag here to say if the
		// start point is the first point is set to true.  If the java code ever
		// treats the last segment differently, change this back to false.		
		Debug.trace( "Segment" + ctrlPs.length );
		singleConnect( ctrlPs[ ctrlPs.length - 1 ], n, true, endP, endN, true, offset );
		Debug.displayTrace();
	}
    wayPoints.push( endP );
    
    Debug.displayTrace( true );
    return Vector.removeRedundant( wayPoints );
}

/**
 * Removes redundant points from a line list (list of points).  A point is 
 * redundant if it colinear with the line segment containing its two neighbours
 * and not outside of that line.  It is also colinear if it is identical to 
 * either of its neighbours.
 *
 * @param points an array of points
 * @returns a new array with redundant points removed.
 */
Vector.removeRedundant = function( points ) {
	var n = points.length;
	var r = new Array();
	if( n == 0 ) return r;
	r.push( points[ 0 ] );
	if( n == 1 ) return r;
	
	
	for( var i = 1; i < n - 1; ++i ) with( Vector ) {
    	var a = points[ i - 1 ], b = points[ i ], c = points[ i + 1 ];
        var ab = sub( b, a ), bc = sub( c, b );

        if( cross( ab, bc ) != 0 || 
        	distanceOrthogonal( a, c ) != distanceOrthogonal( a, b ) + distanceOrthogonal( b, c )) {
            r.push( b );
		}
	}
    if( points[ n - 1 ] != points[ n - 2 ] ) {
    	r.push( points[ n - 1 ] );
	}
	return r;
}    
